*****************************************************************************
*                    Modula-2 Software Construction Set                     *
*                         (c) 1986 by Leon Frenkel                          *
*****************************************************************************
* Name: System.asm                        Version: Amiga.00.00              *
* Created: 11/17/86   Updated: 07/24/87   Author: Leon Frenkel              *
* Description: AmigaDOS Modula-2 program startup code run-time support      *
*   procedures.                                                             *
*****************************************************************************
* Based on: System.OBM (31-Jan-88)        Version: 1.01                     *
* Last Modified: 03-Aug-91                Mods by: Anthon Pang              *
* Description: Added CLI detach code.  See "system.i" for re-assembly info. *
*****************************************************************************

   include "M2/A68K_M2Lib.i" ; NOT "M2Lib.i"--see DATA_BLOCK section at end
   include "m2/mylib.i"      ; my macro library
   include "system.i"        ; module specific constants, macros, etc
;          "m2/system_map.i" ; [see DEFINITION_VARIABLES]

   DEFINITION_PROCEDURES 27
     DEF_PROC MODULE_BODY
     DEF_PROC HALTX
     DEF_PROC MULU32
     DEF_PROC DIVU32
     DEF_PROC MULS32
     DEF_PROC DIVS32
     DEF_PROC FADDs
     DEF_PROC FSUBs
     DEF_PROC FMULs
     DEF_PROC FDIVs
     DEF_PROC FREMs
     DEF_PROC FCMPs
     DEF_PROC FNEGs
     DEF_PROC FABSs
     DEF_PROC FLOATs
     DEF_PROC TRUNCs
     DEF_PROC FADDd
     DEF_PROC FSUBd
     DEF_PROC FMULd
     DEF_PROC FDIVd
     DEF_PROC FREMd
     DEF_PROC FCMPd
     DEF_PROC FNEGd
     DEF_PROC FABSd
     DEF_PROC FLOATd
     DEF_PROC TRUNCd
     DEF_PROC FLONG
     DEF_PROC FSHORT

   DEFINITION_MODULES 0

   DEFINITION_VARIABLES
     include "m2/system_map.i"

   IMPLEMENTATION_VARIABLES

     IFD DETACH

       IFD SAVEDIR
         VAR detach_curdir,ADDRESS
       ENDC

       VAR detach_buffer,ADDRESS
       VAR detach_seglist,ADDRESS
     ENDC

   START_CODE

; -------------------------------------
; Summary of register usage:
;   d0-d1/a0-a1 - scratch variables
;   d2-d7/a2-a6 - temporary
;     d5 - copy of d0 (CmdLineLength)
;     a2 - copy of a0 (CmdLinePtr)
;     a3 - my TaskPtr
;     a4 - module base (for module's local variables)
;     a5 - return address of module initialization table
;     a6 - library base (ExecBase or DOSBase)
;     a7 - stack pointer
; -------------------------------------

   PROCEDURE_CODE MODULE_BODY

*
* Detaching CLI startup code
*
   IFD DETACH

System_ModuleBody:

     lea       MODULE_BASE(pc),a4  ; base for local variables

     move.l    d0,d5          ; length of CLI parameters (temp)
     movea.l   a0,a2          ; pointer to CLI parameters (temp)

     LoadExecBase             ; move.l $0004.w,a6
     suba.l    a1,a1
     jsr       FindTask(a6)   ; get my process structure (extended task structure)
     movea.l   d0,a3          ; d0 = a3 = my_TaskPtr
     tst.l     prCLI(a3)
     beq       From_CLI2      ; branch if CLI detached or WB launched

From_CLI:
                              ; a6 still loaded with ExecBase
     moveq     #LIBRARY_MINIMUM,d0
     lea       DOSName(pc),a1
     jsr       OpenLibrary(a6)
     move.l    d0,Sys_DOSBase(a4)
     beq.s     Fail_Open_DOS

Ok_Open_DOS:
                              ; a6 still loaded with ExecBase
   IFD SAVEDIR
     movea.l   d0,a6
     moveq     #$00,d1
     jsr       CurrentDir(a6) ; cd to \ (return previous "current dir")
     move.l    d0,d1          ; load lock on directory
     move.l    d0,d2          ; save for later
     jsr       CurrentDir(a6) ; cd back to previous "current dir"...ignore result
     move.l    d2,d1          ; duplicate lock for current directory
     jsr       DupLock(a6)
     move.l    d0,detach_curdir(a4)
   ENDC

     movea.l   prCLI(a3),a5    ; pointer to Command Line Interface structure
     B2APTR    a5

     movea.l   cliCommandName(a5),a0
     B2APTR    a0

     move.l    a0,detach_buffer(a4)     ; save ptr to command name (first char is length)

     moveq     #$00,d0
     move.b    (a0),d0        ; length of command name
     move.l    d0,d2          ; save for later

     add.w     d5,d0          ; add length of command line parameters
     addq.w    #$03,d0        ; add for length field & null terminators 
                              ; for substrings

     LoadExecBase
     moveq     #00,d1
     move.l    d0,d3          ; save for later (d3 = size of buffer allocated)
     jsr       AllocMem(a6)

     tst.l     d0
     bne.s     Ok_Alloc_Memory

Fail_Alloc_Memory:

   IFD SAVEDIR
     movea.l   Sys_DOSBase(a4),a6
     move.l    detach_curdir(a4),d1
     jsr       UnLock(a6)

     movea.l   a6,a1
     jsr       CloseLibrary(a6)
   ENDC

   IFND SAVEDIR
     movea.l   Sys_DOSBase(a4),a1
     jsr       CloseLibrary(a6)
   ENDC

Fail_Open_DOS:

     addq      #4,a7          ; pop off return address to module init table

Fail_General:

     moveq     #FAIL,d0
     rts

Ok_Alloc_Memory:
                              ; d0 - contains pointer to allocated block of memory
                              ; d2 - length of command name
                              ; d3 - buffer size
     movea.l   detach_buffer(a4),a0
     movea.l   d0,a1          ; a1 - pointer to command name (source)
     move.l    d0,detach_buffer(a4)
     move.w    d3,(a1)+       ; store size of buffer, in buffer's size field
     move.l    a1,d1          ; save for later (d1 - pointer to BSTR)
     addq.w    #1,d1          ; d1 - pointer to text field

                              ; copy procname length (byte field) from BSTR
                              ; thus we don't decrement d2

CopyName:
                              ; d1 - contains pointer to start of procname string
     move.b    (a0)+,(a1)+
     dbf       d2,CopyName

     clr.b     (a1)+          ; null terminate

     move.l    d5,Sys_CmdLineLength(a4)
     move.l    a1,Sys_CmdLinePtr(a4)

     subq.w    #1,d5

CopyArgs:
                              ; a2 - pointer to CLI args
     move.b    (a2)+,(a1)+
     dbf       d5,CopyArgs

     clr.b     (a1)           ; null terminate

     ; d1 - pointer to null terminated proc name

     moveq     #$00,d2        ; inherit my Task priority
     move.b    Task_tcNode_lnPri(a3),d2

     move.l    cliDefaultStack(a5),d4
     lsl.l     #2,d4          ; inherit my Stack size (in bytes)

     move.l    cliModule(a5),d3
     move.l    d3,detach_seglist(a4)    ; save it

     movea.l   Sys_DOSBase(a4),a6
     jsr       CreateProc(a6)

     tst.l     d0
     bne.s     Ok_Create_Proc

Fail_Create_Proc:

     movea.l   detach_buffer(a4),a1     ; ptr to buffer
     moveq     #0,d0
     move.b    (a1),d0        ; size
     LoadExecBase
     jsr       FreeMem(a6)
     bra.s     Fail_Alloc_Memory

Ok_Create_Proc:

; exit
     addq      #4,a7          ; pop off return address of module init table

     LoadExecBase
     movea.l   Sys_DOSBase(a4),a1
     jsr       CloseLibrary(a6)

     moveq     #$00,d0                  ; return code
     move.l    d0,cliModule(a5)         ; don't free seglist when I exit

     rts


From_CLI2:
                              ; a6 - ExecBase
                              ; a2 - copy of a0 (CLI args ptr)
                              ; d5 - copy of d0 (CLI args length)
                              ; a3 - my TaskPtr
     move.l  a6,Sys_ExecBase(a4)
     lea     OpenLibrary(a6),a5

     Open_Library DOSName,Sys_DOSBase,a5

     Open_Library GraphicsName,Sys_GfxBase,a5

     Open_Library IntuitionName,Sys_IntuitionBase,a5

     Open_Library LayersName,Sys_LayersBase,a5

     Open_Library MathFFPName,Sys_MathBase,a5

     movea.l   (a7)+,a5            ; return address to module init table
     move.l    a7,Sys_StackPtr(a4)

     tst.l     detach_buffer(a4)
     beq       From_WB

  IFD DETACH
     move.l    4(a7),Sys_StackSize(a4)
  ENDC

     move.l    detach_seglist(a4),a2
     B2APTR    a2

     ; count # of segs
     moveq     #1,d0     ; at least 1 for detach_buffer
     move.l    a2,d1     ; store for later

CountNextNode:

     beq.s     FinishCount
     addq.w    #1,d0
     movea.l   d1,a0
     move.l    (a0),d1
     lsl.l     #2,d1
     bra.s     CountNextNode

FinishCount:

     ; d0 = # of mementries required to remember seg nodes
     move.l    d0,d2     ; save for later

     moveq     #$10,d1   ; 16D = size of MemList
     lsl.l     #3,d0     ; multiply by 8...sizeof(struct MemEntry))
     add.l     d1,d0     ; size

     moveq     #0,d1     ; memory type requirements
     jsr       AllocMem(a6)

     tst.l     d0
     beq       Fail_General

     movea.l   d0,a0     ; a0 - ptr to MemList

     move.w    d2,ml_numEntries(a0)

     ; a2 = SegList ptr (APTR)

     lea       ml_ME(a0),a0   ; skip Node & NumEntries field in MemList structure
     move.l    a2,d1

RememberSegs:
     
     beq.s     FinishRemembering
     movea.l   d1,a1
     move.l    (a1),d1        ; get ptr to next seg
     subq      #$4,a1
     move.l    a1,(a0)+       ; store start of seg
     move.l    (a1),(a0)+     ; store size of seg
     lsl.l     #2,d1          ; bptr -> aptr
     bra.s     RememberSegs

FinishRemembering:

     movea.l   detach_buffer(a4),a1
     move.l    a1,(a0)+
     moveq     #00,d1
     move.w    (a1),d1   ; size
     move.l    d1,(a0)

;   a0 points to the tcMemEntry list (in Task structure)
     lea       Task_tcMemEntry(a3),a0 ; get TaskPtr of spawned process
;   d0 is the MemList (contains Node structure)
     movea.l   d0,a1

* code from AddTail macro
     lea       lhTail(a0),a0
     move.l    lnPred(a0),d0
     move.l    a1,lnPred(a0)
     move.l    a0,(a1)
     move.l    d0,lnPred(a1)
     movea.l   d0,a0
     move.l    a1,(a0)

   IFD SAVEDIR
     movea.l   Sys_DOSBase(a4),a6
     move.l    detach_curdir(a4),d1
     jsr       CurrentDir(a6)
   ENDC

   ENDC
*
* End of detaching CLI startup code
*


*
* Normal CLI startup code
*
  IFND DETACH

System_ModuleBody:

     movea.l   (a7)+,a5       ; save return address
     lea       MODULE_BASE(pc),a4
     move.l    a7,Sys_StackPtr(a4)
     move.l    a0,Sys_CmdLinePtr(a4)
     move.l    d0,Sys_CmdLineLength(a4)
     LoadExecBase
     move.l    a6,Sys_ExecBase(a4)

     lea       OpenLibrary(a6),a2

     Open_Library DOSName,Sys_DOSBase,a2

     Open_Library GraphicsName,Sys_GfxBase,a2

     Open_Library IntuitionName,Sys_IntuitionBase,a2

     Open_Library LayersName,Sys_LayersBase,a2

     Open_Library MathFFPName,Sys_MathBase,a2

     suba.l    a1,a1
     jsr       FindTask(a6)   ; my TaskPtr

     movea.l   d0,a3          ; a3 = my TaskPtr
     tst.l     prCLI(a3)      ; my_TaskPtr->prCLI
     beq       From_WB

  ENDC
*
* End of normal CLI startup code
*

*
* Common CLI parse code
*
     suba.w    #$0080,a7
     movea.l   a7,a0
     move.l    a0,Sys_argv(a4)     ; reserve space for 32 longs, argv vectors

     suba.w    #$0100,a7
     movea.l   a7,a1               ; reserve space for parsed CLI arguments

     moveq     #$00,d0
     move.w    #$00fe,d1

  IFND DETACH
     movea.l   prCLI(a3),a2
     B2APTR    a2
     move.l    cliDefaultStack(a2),d2
     lsl.l     #$2,d2
     move.l    d2,Sys_StackSize(a4)
     movea.l   cliCommandName(a2),a2
     B2APTR    a2
  ENDC
  IFD DETACH
     ; no prCLI structure...
     movea.l   detach_buffer(a4),a2
     addq      #$2,a2
  ENDC

     move.l    a1,(a0)+
     addq.w    #$1,d0
     moveq     #$00,d2
     move.b    (a2)+,d2
     subq.w    #$1,d2
1$:
     move.b    (a2)+,(a1)+
     subq.w    #$1,d1
     dbf       d2,1$
     clr.b     (a1)+
     subq.w    #$1,d1
     move.l    Sys_CmdLineLength(a4),d2
     movea.l   Sys_CmdLinePtr(a4),a2
     bra.s     9$

2$:
     cmpi.b    #$20,d7
     bls.s     9$
     cmpi.w    #$0020,d0
     bcc.s     10$
     addq.w    #$1,d0
     move.l    a1,(a0)+
     cmpi.b    #$22,d7
     bne.s     6$
3$:
     move.b    d7,(a1)+
     subq.w    #$1,d1
     bhi.s     4$
     clr.b     (a1)+
     bra.s     10$

4$:
     move.b    (a2)+,d7
     subq.l    #$1,d2
     ble.s     5$
     cmpi.b    #$22,d7
     bne.s     3$
     move.b    d7,(a1)+
     subq.w    #$1,d1
5$:
     clr.b     (a1)+
     subq.w    #$1,d1
     bls.s     10$
     bra.s     9$

6$:
     move.b    d7,(a1)+
     subq.w    #$1,d1
     bhi.s     7$
     clr.b     (a1)+
     bra.s     10$

7$:
     move.b    (a2)+,d7
     subq.l    #$1,d2
     ble.s     8$
     cmpi.b    #$20,d7
     bhi.s     6$
8$:
     clr.b     (a1)+
     subq.w    #$1,d1
     bls.s     10$
9$:
     move.b    (a2)+,d7
     subq.l    #$1,d2
     bgt.s     2$
10$:
     move.w    d0,Sys_argc(a4)
     movea.l   Sys_DOSBase(a4),a6
     jsr       Input(a6)
     move.l    d0,Sys_StdInput(a4)
     jsr       Output(a6)
     move.l    d0,Sys_StdOutput(a4)
     bra.s     EndStartup


*
* Start of WB startup code
*
From_WB:

  IFD DETACH
     move.l    a2,Sys_CmdLinePtr(a4)
     move.l    d5,Sys_CmdLineLength(a4)
  ENDC

     move.l    prStackSize(a3),Sys_StackSize(a4)
     lea       prMsgPort(a3),a0
     jsr       WaitPort(a6)        ; wait for Workbench msg
     lea       prMsgPort(a3),a0
     jsr       GetMsg(a6)          ; get Workbench msg
     move.l    d0,Sys_WBenchMsg(a4)
     movea.l   d0,a2
     movea.l   Sys_DOSBase(a4),a6
                                   ; a2 - pointer to WBenchMsg
     movea.l   wbArgPtr(a2),a2     ; a2 - pointer to WBArg structure
     move.l    waLock(a2),d1
     jsr       CurrentDir(a6)

     movea.l   waName(a2),a0

     move.l    a7,d7                    ; d7 = save stack ptr (top)
     suba.w    #$0100,a7                ; allocate 256 bytes for buffer
     movea.l   a7,a1                    ; a1 = base of buffer (curr_pointer)
                                        ; bottom (destination start)

     lea       RAWWindowName(pc),a2     ; copy window params to buffer
     moveq     #RAWWindowEnd-RAWWindowName-1,d0   ; length
CopyWindowName:
     move.b    (a2)+,(a1)+
     dbf       d0,CopyWindowName

     moveq     #$7f,d0                  ; append program name
AppendProcName:
     move.b    (a0)+,(a1)+
     dbeq      d0,AppendProcName        ; until -1 or null reached
     clr.b     (a1)+                    ; null terminate

     move.l    a7,d1                    ; buffer base
     move.l    #ModeOldFile,d2
     jsr       Open(a6)                 ; open RAW window
     move.l    d0,Sys_StdInput(a4)
     move.l    d0,Sys_StdOutput(a4)
     lsl.l     #$2,d0                   ; convert to APTR
     movea.l   d0,a0                    ; filehandle
     move.l    fhType(a0),prConsoleTask(a3)
     movea.l   d7,a7                    ; restore stack ptr

EndStartup:

     jsr       (a5)
*
* End of WB startup code
*


*
* Exit code
*
   PROCEDURE_CODE HALTX

System_HALTX:

     movea.l   MODULE_BASE+Sys_StackPtr(pc),a7
     move.l    MODULE_BASE+Sys_WBenchMsg(pc),d2
     beq.s     HALT_From_CLI

     movea.l   MODULE_BASE+Sys_DOSBase(pc),a6
     move.l    MODULE_BASE+Sys_StdOutput(pc),d1
     jsr       Close(a6)

     LoadExecBase
     jsr       Forbid(a6)
     movea.l   d2,a1
     jsr       ReplyMsg(a6)

HALT_From_CLI:

  IFD DETACH
     IFD SAVEDIR
       movea.l   MODULE_BASE+Sys_DOSBase(pc),a6
       move.l    MODULE_BASE+detach_curdir(pc),d1
       jsr       UnLock(a6)
       movea.l   a6,a1
     ENDC
     IFND SAVEDIR
       movea.l    MODULE_BASE+Sys_DOSBase(pc),a1
     ENDC
  ENDC
  IFND DETACH
     movea.l   MODULE_BASE+Sys_DOSBase(pc),a1
  ENDC

     LoadExecBase
     lea       CloseLibrary(a6),a2

     ; a1 loaded with Sys_DOSBase
     jsr       (a2)

     movea.l   MODULE_BASE+Sys_GfxBase(pc),a1
     jsr       (a2)

     movea.l   MODULE_BASE+Sys_IntuitionBase(pc),a1
     jsr       (a2)

     movea.l   MODULE_BASE+Sys_LayersBase(pc),a1
     jsr       (a2)

     movea.l   MODULE_BASE+Sys_MathBase(pc),a1
     jsr       (a2)

     move.l    MODULE_BASE+Sys_CLIReturnCode(pc),d0
     rts

; -------------------------------------

*
* string constants
*
DOSName:       dc.b 'dos.library',$00

GraphicsName:  dc.b 'graphics.library',$00

IntuitionName: dc.b 'intuition.library',$00

LayersName:    dc.b 'layers.library',$00

MathFFPName:   dc.b 'mathffp.library',$00

RAWWindowName: dc.b 'RAW:10/10/320/80/'
RAWWindowEnd:

     ds.w 0    ; alignment check

; -------------------------------------

   PROCEDURE_CODE MULU32 ; System_MULU32

; ----------

   PROCEDURE_CODE MULS32 ; System_MULS32

     movem.l   d2-d3,-(a7)
     move.w    d1,d2
     mulu      d0,d2
     move.l    d1,d3
     swap      d3
     mulu      d0,d3
     swap      d3
     clr.w     d3
     add.l     d3,d2
     swap      d0
     mulu      d1,d0
     swap      d0
     clr.w     d0
     add.l     d2,d0
     movem.l   (a7)+,d2-d3
     rts

; ----------

   PROCEDURE_CODE DIVU32 ; System_DIVU32

     move.l    d2,-(a7)
     move.l    d1,d2
     move.l    d0,d1

DIV32PartialQuotients:

     cmpi.l    #$0000ffff,d2
     bgt.s     DIV32ShiftSubtract

     movea.w   d1,a6
     clr.w     d1
     swap      d1
     divu      d2,d1
     move.l    d1,d0
     swap      d1
     move.w    a6,d0
     divu      d2,d0
     move.w    d0,d1
     clr.w     d0
     swap      d0
     exg       d0,d1
     move.l    (a7)+,d2
     rts

DIV32ShiftSubtract:

     move.l    d1,d0
     clr.w     d0
     swap      d0
     swap      d1
     clr.w     d1
     movea.l   d2,a6
     moveq     #$0f,d2
1$:
     add.l     d1,d1
     addx.l    d0,d0
     cmpa.l    d0,a6
     bgt.s     2$
     sub.l     a6,d0
     addq.w    #$1,d1
2$:
     dbf       d2,1$
     exg       d0,d1
     move.l    (a7)+,d2
     rts

; ----------

   PROCEDURE_CODE DIVS32 ; System_DIVS32

     move.w    d3,-(a7)
     pea       2$(pc)
     move.l    d2,-(a7)
     move.l    d1,d2
     smi       d3
     bpl.s     1$
     neg.l     d2
1$:
     move.l    d0,d1
     bpl.s     DIV32PartialQuotients
     neg.l     d1
     not.b     d3
     bra.s     DIV32PartialQuotients

2$:
     tst.b     d3
     beq.s     3$
     neg.l     d0
3$:
     move.w    (a7)+,d3
     rts

; ----------

   EMPTY_STUB FADDd ; System_FADDd

; ----------

   EMPTY_STUB FSUBd ; System_FSUBd

; ----------

   EMPTY_STUB FMULd ; System_FMULd

; ----------

   EMPTY_STUB FDIVd ; System_FDIVd

; ----------

   EMPTY_STUB FREMd ; System_FREMd

; ----------

   EMPTY_STUB FCMPd ; System_FCMPd

; ----------

   EMPTY_STUB FNEGd ; System_FNEGd

; ----------

   EMPTY_STUB FABSd ; System_FABSd

; ----------

   EMPTY_STUB FLOATd ; System_FLOATd

; ----------

   EMPTY_STUB TRUNCd ; System_TRUNCd

; ----------

   EMPTY_STUB FLONG ; System_FLONG

; ----------

   EMPTY_STUB FSHORT ; System_FSHORT

; ----------

   DOUBLE_PARAM_STUB FADDs,-$0042 ; System_FADDs

; ----------

   DOUBLE_PARAM_STUB FSUBs,-$0048 ; System_FSUBs

; ----------

   DOUBLE_PARAM_STUB FMULs,-$004e ; System_MULs

; ----------

   DOUBLE_PARAM_STUB FDIVs,-$0054 ; System_FDIVs

; ----------

   EMPTY_STUB FREMs ; System_FREMs

; ----------

   DOUBLE_PARAM_STUB FCMPs,-$002a ; System_FCMPs

; ----------

   SINGLE_PARAM_STUB FNEGs,-$003c ; System_FNEGs

; ----------

   SINGLE_PARAM_STUB FABSs,-$0036 ; System_FABSs

; ----------

   SINGLE_PARAM_STUB FLOATs,-$0024 ; System_FLOATs

; ----------

   SINGLE_PARAM_STUB TRUNCs,-$001e ; System_TRUNCs

; -------------------------------------

   END_CODE

*
* Offsets to routines must be determined in the same order as the
* procedures appear in System.def. Since the code relies on routines
* to fall through in certain places, we must calculate these offsets
* independent from the order it is written.  Therefore, you must use
* A68K_M2Lib.i, instead of M2Lib.i.
*
     SECTION HEADER_BLOCK,DATA
       dc.W CODE_BLOCK_END-CODE_BLOCK_START
       dc.W DEF_VAR_SIZE
       dc.W IMP_VAR_SIZE
       dc.W DEF_PROCEDURE_COUNT
       dc.W DEF_MODULE_COUNT

     SECTION DATA_BLOCK,DATA
       dc.w _MODULE_BODY-MODULE_BASE
       dc.w _HALTX-MODULE_BASE
       dc.w _MULU32-MODULE_BASE
       dc.w _DIVU32-MODULE_BASE
       dc.w _MULS32-MODULE_BASE
       dc.w _DIVS32-MODULE_BASE
       dc.w _FADDs-MODULE_BASE
       dc.w _FSUBs-MODULE_BASE
       dc.w _FMULs-MODULE_BASE
       dc.w _FDIVs-MODULE_BASE
       dc.w _FREMs-MODULE_BASE
       dc.w _FCMPs-MODULE_BASE
       dc.w _FNEGs-MODULE_BASE
       dc.w _FABSs-MODULE_BASE
       dc.w _FLOATs-MODULE_BASE
       dc.w _TRUNCs-MODULE_BASE
       dc.w _FADDd-MODULE_BASE
       dc.w _FSUBd-MODULE_BASE
       dc.w _FMULd-MODULE_BASE
       dc.w _FDIVd-MODULE_BASE
       dc.w _FREMd-MODULE_BASE
       dc.w _FCMPd-MODULE_BASE
       dc.w _FNEGd-MODULE_BASE
       dc.w _FABSd-MODULE_BASE
       dc.w _FLOATd-MODULE_BASE
       dc.w _TRUNCd-MODULE_BASE
       dc.w _FLONG-MODULE_BASE
       dc.w _FSHORT-MODULE_BASE

   END
